/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#ifndef fmr_ens
#define fmr_ens 	1
#ifndef ELEMENTMAX
#define ELEMENTMAX     5001
#endif

/*=64 en archi 64 bits */
#define NB_BITS        (unsigned int)(sizeof(long)*8)        
/* arrondi par valeur sup*/
#define NB_LONG(m)        (unsigned int)(1+ (m)/NB_BITS) 

typedef unsigned int Element_t;
typedef unsigned long * Ensemble_t;

bool 
  	Ens_appartient(Element_t,Ensemble_t), /* elt appartient a ensemble ?*/
	Ens_est_vide(Ensemble_t, unsigned int),        /* ensemble vide ? */
	Ens_est_singleton(Ensemble_t, unsigned int );  /* ensemble réduit à un seul elt ? */

Ensemble_t
	Ens_nouveau(unsigned int maxelt); 		  /* cree un nouvel ensemble en param taille max elt*/

Element_t
        Ens_elt_suivant(Element_t el, Ensemble_t s, unsigned int maxelt); /* trouve l'element suivant element dans l'ens.*/
unsigned int
	Ens_ote_prem_elt(Ensemble_t set, unsigned int maxelt); /* trouve le premier elt, renvoie -1 si ens vide*/



void
  	Ens_ajoute(Element_t,Ensemble_t),    /* Ajoute un elt a un ens. */
  	Ens_ote(Element_t,Ensemble_t),       /* Ote un elt a un ens. */
	Ens_rempli(Ensemble_t set, unsigned int maxelt), /* ajoute les éléments de n à m inclus dans l'ens. */
	Ens_zero(Ensemble_t, unsigned int),		/* reinit. un ens. */
	Ens_copy(Ensemble_t ,const Ensemble_t, unsigned int maxelt);

unsigned
	Ens_aff(Ensemble_t, unsigned int),              /* Affiche un ensemble par ses elements,
                                      	retourne le cardinal */
	Ens_appli(), 			/* Applique une fonc. user a chaque elt. d'un 
                                      	ens. et retourne le cardinal*/
        Ens_card(Ensemble_t, unsigned int);              /* retourne le cardinal de set (plus rapide que 
                                      	l'utilisation de Ens_appli)*/

#endif //fmr_ens
