/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>

#include "fmr.h"
#include "fmr_mem.h"
#include "fmr_constraints.h"
#include "fmr_cost.h"
#include "fmr_psmvt.h"
#include "fmr_exploShift.h"
#include "fmr_heap.h"
#include "fmr_threads.h"

/**
 * Algo qui explore un voisinage par shift (distance 1), on ne garde que le meilleur déplacement
 * valide et ont le fait si améliorant.
 */
long
explo_voisinageVeryBestShift(const placement_t const *initial, placement_t *resultat, long gainMin) {
    unsigned short p,m_orig,m, nbShifts=0, nbBonsShifts=0 ;
    unsigned short p_best=NOPROCESS, m_best=NOMACHINE;
    long gain, gainMax=gainMin;

        for (p=0; p< instance.nb_procs; p++) {
    for (m=0; m< instance.nb_machines; m++) {
	   m_orig= resultat->assignment[p];
	   if (m !=m_orig && check_all_constraints_shift( resultat, p , m_orig, m)) {
		gain=calcule_gain_shift(initial, resultat, p, m_orig, m);
		if (gain>gainMax) {
			gainMax=gain;
			p_best=p;
			m_best=m;
		}
	   }
	}
    }

    if (p_best==NOPROCESS)
	return(0);
    else {
     	  deplace_proc(initial, resultat, p_best , m_best); 
	  resultat->cout -= gainMax ;
          DEBUG_PRINT("deplace %4d (service %3d) de %3d à %3d, gain=%ld\n", p_best, instance.procs[p_best].service, initial->assignment[p_best], m_best, gainMax);
	  return(gainMax);
    }
}

long
explo_voisinageVeryBestShift_restreint(const placement_t const *initial, placement_t *resultat, unsigned short present_ou_indice[], long gainMin) {
    unsigned short p,m_orig,m, nbShifts=0, nbBonsShifts=0, indice ;
    unsigned short p_best=NOPROCESS, m_best=NOMACHINE;
    long gain, gainMax=gainMin;

    for (p=0; p< instance.nb_procs; p++) {
      	if ((indice=present_ou_indice[p])!=NOPROCESS) { // ps oté par le glouton partiel, on saute à l'indice suivant
	      	if (indice ==0) { // il n'y a plus de ps non oté
		      	break;
			} else 
			p=indice;
		} // sinon il est bien présent.
    for (m=0; m< instance.nb_machines; m++) {

	   m_orig= resultat->assignment[p];
	   if (m !=m_orig && check_all_constraints_shift( resultat, p , m_orig, m)) {
		gain=calcule_gain_shift(initial, resultat, p, m_orig, m);
		if (gain>gainMax) {
			gainMax=gain;
			p_best=p;
			m_best=m;
		}
	   }
	}
    }

    if (p_best==NOPROCESS)
	return(0);
    else {
     	  deplace_proc(initial, resultat, p_best , m_best); 
	  resultat->cout -= gainMax ;
          DEBUG_PRINT("deplace %4d (service %3d) de %3d à %3d, gain=%ld\n", p_best, instance.procs[p_best].service, initial->assignment[p_best], m_best, gainMax);
	  return(gainMax);
    }
}

/**
 * Algo qui explore un voisinage par shift (distance 1), on ne garde que les déplacemnts
 * valides dans une Heap et à la fin on ne fait que le déplacement du meilleur (c'est un essai!)
 */
long
explo_voisinageBestShift(const placement_t const *initial, placement_t *resultat, long gainMin) {
   //placement_t pl_temp;
   unsigned short p,m_orig,m,i ;
   long gain, gainTotal=0;

	/* Arbre binaire quasi complet pour stocker les meilleurs mouvements,
	   seul le loadcost est considéré (semble le critère prépondérant).
	   tabGain sera donc ordonné */
   unsigned nbEltTabGain=0, perdus=0 ;
   deplt_t **tabGain;

   if ((unsigned long) instance.nb_procs*instance.nb_machines > (unsigned long)MAXNBDEPL)
	tabGain=(deplt_t **)malloc(sizeof(deplt_t *)*MAXNBDEPL);
   else
	tabGain=(deplt_t **)malloc(sizeof(deplt_t *)*instance.nb_procs*instance.nb_machines);

   if (tabGain==NULL ) {
	   DEBUG_PRINT("malloc returns NULL\n");
	   return(0L);
   }
   bool *dejaVu= (bool *) calloc (instance.nb_procs, sizeof(bool));

    for (p=0; p< instance.nb_procs; p++) {
	/*pthread_testcancel();*/

	m_orig= resultat->assignment[p];
	for (m=0; m< instance.nb_machines; m++) {
	/* on ne garde que les deplacements valides */
	   if (m !=m_orig && check_all_constraints_shift( resultat, p , m_orig, m)) {
		if ( (gain=calcule_gain_shift(initial, resultat, p, m_orig, m)) >gainMin)
		  ajoutHeap( p,m, gain , tabGain, &nbEltTabGain);
	   }
	}
     }
     //DEBUG_PRINT("Nbre de deplacement shift trouvés:%u\n", nbEltTabGain);
     DEBUG_STATS_PRINT("Nbre de deplacement shift trouvés:%u/%u\n", nbEltTabGain, MAXNBDEPL);

	/* on traite spécifiquement le meilleur car son gain est exact */
     if (nbEltTabGain>0 && tabGain[0]->gainLoad >gainMin) {
          DEBUG_PRINT("depl de %u sur %u à %u, gain=%ld\n", tabGain[0]->u1 , resultat->assignment[tabGain[0]->u1], tabGain[0]->u2, tabGain[0]->gainLoad);
     	  deplace_proc(initial, resultat, tabGain[0]->u1 , tabGain[0]->u2); 
	  gainTotal +=tabGain[0]->gainLoad;
	  dejaVu[tabGain[0]->u1]= true;
	  atomic_set_meilleur_placement(resultat,'b');
     } else {
	  DEBUG_PRINT("meme pas une amélioration!\n");
	  DEBUG_STATS_PRINT("*** meme pas une amélioration: %u entreees de heap perdues\n",nbEltTabGain);
	  free_heap_elts(tabGain, nbEltTabGain);
	  free(tabGain);
	  return FMR_OK; /* pas la peine d'aller plus loin */
     }

	/* les gains suivants sont estimés mais tant pis */
     while(nbEltTabGain>0 && tabGain[0]->gainLoad> gainMin) {
	m_orig=resultat->assignment[tabGain[0]->u1];
	if (! dejaVu[tabGain[0]->u1] 
		&& check_all_constraints_shift(resultat, tabGain[0]->u1, m_orig, tabGain[0]->u2)
		&& (gain=calcule_gain_shift(initial, resultat, tabGain[0]->u1, m_orig, tabGain[0]->u2))>gainMin) {
		DEBUG_PRINT("depl de %u de %u à %u, gain = %ld\n", tabGain[0]->u1 ,
                          resultat->assignment[tabGain[0]->u1], tabGain[0]->u2, gain);
                gainTotal +=gain;
                deplace_proc(initial, resultat, tabGain[0]->u1 , tabGain[0]->u2);
		atomic_set_meilleur_placement(resultat,'b');
		dejaVu[tabGain[0]->u1]= true;
	} else {
		perdus++;
	}

	oteHeap(tabGain, & nbEltTabGain);
     }

     DEBUG_STATS_PRINT("nbre perdus=%u\n",perdus);
     resultat->cout -= gainTotal ;
     DEBUG_PRINT("Gain total explo_voisinageBestShift:%ld\n",gainTotal);
     free_heap_elts(tabGain, nbEltTabGain);
     free(tabGain);
     free(dejaVu);
     return gainTotal;
}

long
explo_voisinageBestShift_restreint(const placement_t const *initial, placement_t *resultat, unsigned short present_ou_indice[], long gainMin) {
   //placement_t pl_temp;
   unsigned short p,m_orig,m,i, indice ;
   long gain, gainTotal=0;

	/* Arbre binaire quasi complet pour stocker les meilleurs mouvements,
	   seul le loadcost est considéré (semble le critère prépondérant).
	   tabGain sera donc ordonné */
   unsigned nbEltTabGain=0 ;
   deplt_t **tabGain;

   if ((unsigned long) instance.nb_procs*instance.nb_machines > (unsigned long)MAXNBDEPL)
	tabGain=(deplt_t **)malloc(sizeof(deplt_t *)*MAXNBDEPL);
   else
	tabGain=(deplt_t **)malloc(sizeof(deplt_t *)*instance.nb_procs*instance.nb_machines);

   if (tabGain==NULL ) {
	   DEBUG_PRINT("malloc returns NULL\n");
	   return(0L);
   }
   bool *dejaVu= (bool *) calloc (instance.nb_procs, sizeof(bool));

    for (p=0; p< instance.nb_procs; p++) {
	if ((indice=present_ou_indice[p])!=NOPROCESS) { // ps oté par le glouton partiel, on saute à l'indice suivant
		if (indice ==0) { // il n'y a plus de ps non oté
			break;
		} else 
			p=indice;
	} // sinon il est bien présent.

	m_orig= resultat->assignment[p];
	for (m=0; m< instance.nb_machines; m++) {
	/* on ne garde que les deplacements valides */
	   if (m !=m_orig && check_all_constraints_shift( resultat, p , m_orig, m)) {
		if ( (gain=calcule_gain_shift(initial, resultat, p, m_orig, m)) >gainMin)
		  ajoutHeap( p,m, gain , tabGain, &nbEltTabGain);
	   }
	}
     }
     DEBUG_PRINT("Nbre de deplacement shift trouvés:%u\n", nbEltTabGain);

	/* on traite spécifiquement le meilleur car son gain est exact */
     if (nbEltTabGain>0 && tabGain[0]->gainLoad >gainMin) {
          DEBUG_PRINT("depl de %u sur %u à %u, gain=%ld\n", tabGain[0]->u1 , resultat->assignment[tabGain[0]->u1], tabGain[0]->u2, tabGain[0]->gainLoad);
     	  deplace_proc(initial, resultat, tabGain[0]->u1 , tabGain[0]->u2); 
	  gainTotal +=tabGain[0]->gainLoad;
	  dejaVu[tabGain[0]->u1]= true;
     } else {
	  DEBUG_PRINT("meme pas une amélioration!\n");
	  free_heap_elts(tabGain, nbEltTabGain);
	  free(tabGain);
	  return FMR_OK; /* pas la peine d'aller plus loin */
     }

	/* les gains suivants sont estimés mais tant pis */
     while(nbEltTabGain>0 && tabGain[0]->gainLoad> gainMin) {
	m_orig=resultat->assignment[tabGain[0]->u1];
	if (! dejaVu[tabGain[0]->u1] 
		&& check_all_constraints_shift(resultat, tabGain[0]->u1, m_orig, tabGain[0]->u2)
		&& (gain=calcule_gain_shift(initial, resultat, tabGain[0]->u1, m_orig, tabGain[0]->u2))>gainMin) {
		DEBUG_PRINT("depl de %u de %u à %u, gain = %ld\n", tabGain[0]->u1 ,
                          resultat->assignment[tabGain[0]->u1], tabGain[0]->u2, gain);
                gainTotal +=gain;
                deplace_proc(initial, resultat, tabGain[0]->u1 , tabGain[0]->u2);
		dejaVu[tabGain[0]->u1]= true;
	}
	oteHeap(tabGain, & nbEltTabGain);
     }

     resultat->cout -= gainTotal ;
     DEBUG_PRINT("Gain total explo_voisinageBestShift:%ld\n",gainTotal);
     free_heap_elts(tabGain, nbEltTabGain);
     free(tabGain);
     free(dejaVu);
     return gainTotal;
}

/**
 * Algo qui explore un voisinage par shift (distance 1), on ne garde que les déplacemnts
 * valides et ont les fait tout de suite si améliorants.
 */
long
explo_voisinageShift(const placement_t const *initial, placement_t* resultat, long gainMin) {
    unsigned short p,m_orig,m, nbShifts=0, nbBonsShifts=0 ;
    long gain, gainTotal=0;
    
    for (m=0; m< instance.nb_machines; m++) {
        for (p=0; p< instance.nb_procs; p++) {
	    /*pthread_testcancel();*/

            m_orig= resultat->assignment[p];
            /* on ne garde que les deplacements valides */
            if (m != m_orig && check_all_constraints_shift( resultat, p , m_orig, m)) {
                nbShifts++;
                if ( (gain = calcule_gain_shift (initial, resultat, p, m_orig, m)) >gainMin) {
                    nbBonsShifts++;
                    gainTotal +=gain;
                    deplace_proc(initial, resultat, p, m);
                    //DEBUG_PRINT("deplace %4d (service %3d) de %3d à %3d, gain=%ld\n", p, instance.procs[p].service, m_orig, m, gain);
		    atomic_set_meilleur_placement(resultat,'s');
                }
            }
        }
    }
    resultat->cout -= gainTotal ;
    DEBUG_PRINT("Nbre de bon deplacements (shift) trouvés:%u/%u valides\n", nbBonsShifts, nbShifts);
    DEBUG_PRINT("Gain total explo_voisinageShift:%ld\n",gainTotal);
    return gainTotal;
}

long
explo_voisinageShift2(const placement_t const *initial, placement_t* resultat, long gainMin) {
    unsigned short p,m_orig,m, nbShifts=0, nbBonsShifts=0 ;
    long gain, gainTotal=0;
    
        for (p=0; p< instance.nb_procs; p++) {
    for (m=0; m< instance.nb_machines; m++) {

            m_orig= resultat->assignment[p];
            /* on ne garde que les deplacements valides */
            if (m != m_orig && check_all_constraints_shift( resultat, p , m_orig, m)) {
                nbShifts++;
                if ( (gain = calcule_gain_shift (initial, resultat, p, m_orig, m)) >gainMin) {
                    nbBonsShifts++;
                    gainTotal +=gain;
                    deplace_proc(initial, resultat, p, m);
                    //DEBUG_PRINT("deplace %4d (service %3d) de %3d à %3d, gain=%ld\n", p, instance.procs[p].service, m_orig, m, gain);
		    atomic_set_meilleur_placement(resultat,'z');
                }
            }
        }
    }
    resultat->cout -= gainTotal ;
    DEBUG_PRINT("Nbre de bon deplacements (shift) trouvés:%u/%u valides\n", nbBonsShifts, nbShifts);
    DEBUG_PRINT("Gain total explo_voisinageShift:%ld\n",gainTotal);
    return gainTotal;
}

long
explo_voisinageShift_restreint(const placement_t const *initial, placement_t* resultat, unsigned short present_ou_indice[], long gainMin) {
    unsigned short p,m_orig,m, nbShifts=0, nbBonsShifts=0, indice ;
    long gain, gainTotal=0;
    
    for (p=0; p< instance.nb_procs; p++) {
	    if ((indice=present_ou_indice[p])!=NOPROCESS) { // ps oté par le glouton partiel, on saute à l'indice suivant
		if (indice ==0) { // il n'y a plus de ps non oté
			break;
		} else 
			p=indice;
	    } // sinon il est bien présent.

    for (m=0; m< instance.nb_machines; m++) {
            m_orig= resultat->assignment[p];
            /* on ne garde que les deplacements valides */
            if (m != m_orig && check_all_constraints_shift( resultat, p , m_orig, m)) {
                nbShifts++;
                if ( (gain = calcule_gain_shift (initial, resultat, p, m_orig, m)) >gainMin) {
                    nbBonsShifts++;
                    gainTotal +=gain;
                    deplace_proc(initial, resultat, p, m);
                    //DEBUG_PRINT("deplace %4d (service %3d) de %3d à %3d, gain=%ld\n", p, instance.procs[p].service, m_orig, m, gain);
                }
            }
        }
    }
    resultat->cout -= gainTotal ;
    DEBUG_PRINT("Nbre de bon deplacements (shift) trouvés:%u/%u valides\n", nbBonsShifts, nbShifts);
    DEBUG_PRINT("Gain total explo_voisinageShift:%ld\n",gainTotal);
    return gainTotal;
}
