/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include "fmr_profondeur.h"

#include "fmr.h"
#include "fmr_mem.h"
#include "fmr_cost.h"
#include "fmr_tools.h"
#include "fmr_exploRandSwap.h"
#include "fmr_exploShift.h"
#include "fmr_exploSwap.h"
#include "fmr_exploEjection.h"
#include "fmr_vns.h"
#include "fmr_recuit.h"
#include "fmr_recuit_b.h"
#include "fmr_glouton.h"
#include "fmr_core.h"
#include "fmr_threads.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>


/**
 * Le calcul lui-meme.
 * Prend en entree un pointeur vers l'instance, un pointeur vers
 * le placement initial et un pointeur vers le resultat du calcul.
 * Retourne FMR_OK si tout s'est bien passe, une erreur si il 
 * y a eu un probleme.
 */

int
calcul_assignement_rec (const placement_t * initial, placement_t * courant) {
    int64 gain=0;
    long gainMin = 0;		/* a parametrer dynamiquement */
    unsigned short longChaine = MAXLONGCHAINEEJECTION;	/* choisi par experimentation */
    placement_t resultat_courant;
    placement_t meilleur_resultat;

    char * listeAlgos = "XBCWYES";
//        char * listeAlgos = "BW";
    
    resultat_courant = alloc_assignement(instance.nb_procs);
    meilleur_resultat = alloc_assignement(instance.nb_procs);
    copier_assignement( courant, &meilleur_resultat );
    
    for (; *listeAlgos!=0; listeAlgos++) {

	    copier_assignement(courant, &resultat_courant);

	    switch (*listeAlgos) {
            case 'E':
                gain = explo_voisinageSwap (initial, &resultat_courant, gainMin);	//Echange
                break;
            case 'X':
                gain = explo_voisinageVeryBestSwap(initial, &resultat_courant, gainMin);	//Echange
                break;
            case 'W':
                gain = explo_voisinageBestSwap(initial, &resultat_courant, gainMin);	//Echange
                break;
            case 'S':
                gain = explo_voisinageShift (initial, &resultat_courant, gainMin);
                break;
            case 'B':
                gain = explo_voisinageBestShift (initial, &resultat_courant, gainMin);	//Best shift
                break;
            case 'Y':
                gain = explo_voisinageVeryBestShift (initial, &resultat_courant, gainMin);// Very Best shift
                break;
            case 'C':
                gain = explo_voisinageEjectionChain (initial, &resultat_courant, longChaine,
                                                   gainMin);
                break;
            default:
                fprintf (stderr, "Erreur algo inconnu\n");
		gain = 0;
                break;
        }
        if (gain>gainMin) {
	    atomic_set_meilleur_placement(&resultat_courant, *listeAlgos);
	    gain += calcul_assignement_rec(initial, &resultat_courant) ;
	    if (resultat_courant.cout < meilleur_resultat.cout)
		copier_assignement(&resultat_courant, &meilleur_resultat);
	}
    }
    if (alarme)
        return FMR_TIMEOUT;
    
    gain = courant->cout - meilleur_resultat.cout;
    copier_assignement(&meilleur_resultat, courant);
    free_placement( &resultat_courant );
    free_placement( &meilleur_resultat);
    return (int) gain;
}

int
calcul_assignement_rec_max (const placement_t * initial, placement_t * courant, int profmax) {
    int64 gain=0;
    long gainMin = 0;		/* a parametrer dynamiquement */
    unsigned short longChaine = MAXLONGCHAINEEJECTION;	/* choisi par experimentation */
    placement_t resultat_courant;
    placement_t meilleur_resultat;
    char * listeAlgos = "BWCSEYX";
//        char * listeAlgos = "BW";
    
    if (profmax==0) return 0;

    meilleur_resultat = alloc_assignement(instance.nb_procs);
    resultat_courant = alloc_assignement(instance.nb_procs);
    
    
    copier_assignement( courant, &meilleur_resultat );
    for (; *listeAlgos!=0; listeAlgos++) {

	    copier_assignement(courant, &resultat_courant);

	    switch (*listeAlgos) {
            case 'E':
                gain = explo_voisinageSwap (initial, &resultat_courant, gainMin);	//Echange
                break;
            case 'X':
                gain = explo_voisinageVeryBestSwap(initial, &resultat_courant, gainMin);	//Echange
                break;
            case 'W':
                gain = explo_voisinageBestSwap(initial, &resultat_courant, gainMin);	//Echange
                break;
            case 'S':
                gain = explo_voisinageShift (initial, &resultat_courant, gainMin);
                break;
            case 'B':
                gain = explo_voisinageBestShift (initial, &resultat_courant, gainMin);	//Best shift
                break;
            case 'Y':
                gain = explo_voisinageVeryBestShift (initial, &resultat_courant, gainMin);// Very Best shift
                break;
            case 'C':
                gain = explo_voisinageEjectionChain (initial, &resultat_courant, longChaine,
                                                   gainMin);
                break;
            default:
                fprintf (stderr, "Erreur algo inconnu\n");
		gain = 0;
                break;
        }
        if (gain>gainMin) {
	    atomic_set_meilleur_placement(&resultat_courant, *listeAlgos);
	    gain += calcul_assignement_rec_max(initial, &resultat_courant, profmax-1) ;
	    if (resultat_courant.cout < meilleur_resultat.cout)
		copier_assignement(&resultat_courant, &meilleur_resultat);
	}
    }
    if (alarme)
        return FMR_TIMEOUT;
    
    gain = courant->cout - meilleur_resultat.cout;
    copier_assignement(&meilleur_resultat, courant);
    free_placement( &resultat_courant );
    free_placement( &meilleur_resultat);
    return (int) gain;
}

